/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.layout.auto;

import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodePosition;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setStickyConnectorLayoutData;

import java.util.Map.Entry;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.BasicEMap;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.EMap;
import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.base.ExitConnectorBase;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Points;
import org.fortiss.tooling.base.model.layout.impl.PointsImpl;
import org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase;
import org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils;
import org.fortiss.tooling.base.utils.LayoutDataUtils;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

import de.cau.cs.kieler.core.alg.BasicProgressMonitor;
import de.cau.cs.kieler.core.alg.IKielerProgressMonitor;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.AbstractLayoutProvider;
import de.cau.cs.kieler.kiml.klayoutdata.KEdgeLayout;
import de.cau.cs.kieler.kiml.klayoutdata.KPoint;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.Direction;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.PortSide;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klay.layered.LayeredLayoutProvider;
import de.cau.cs.kieler.klay.layered.properties.InteractiveReferencePoint;
import de.cau.cs.kieler.klay.layered.properties.Properties;

/**
 * <p>
 * {@link IAutoLayouter} based on <a
 * href="https://rtsys.informatik.uni-kiel.de/confluence/display/KIELER/Home">KIELER framework</a>.
 * </p>
 * 
 * <
 * 
 * Approach:
 * <ul>
 * <li>Generate a KIELER graph from the input model</li>
 * <li>Apply the KIELER <i>layered layout algorithm</i></li>
 * <li>Transfer generated layout to input model
 * </ul>
 * 
 * @author offtermatt, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 8BE55A0C36B936EF103CD3350B857042
 */
public class KielerAutoLayouter implements IAutoLayouter {

	/** Mapping from {@link IHierarchicElement}s to {@link KNode}s. */
	private EMap<IHierarchicElement, KNode> modelElementsToKNodes;

	/** Mapping from outbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> outboundConnectorsToKPorts;

	/** Mapping from inbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> inboundConnectorsToKPorts;

	/** Used to map undirected {@link IConnector}s to {@link KPort}s */
	private EMap<IConnector, KPort> undirectedConnectorsToKPorts;

	/** Mapping from {@link IConnection}s to {@link KEdge}s. */
	private EMap<IConnection, KEdge> connectionsToKEdges;

	/** Mapping from global inbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> globalInboundConnectorsToKPorts;

	/** Mapping from global outbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> globalOutboundConnectorsToKPorts;

	/** {@inheritDoc} */
	@Override
	public void performAutoLayout(IHierarchicElement hierarchicElement) {

		boolean avUpdateEnabled = AnnotationViewPartBase.isUpdateEnabled();
		AnnotationViewPartBase.setUpdateEnabled(false);
		KNode parentNode = createKIELERGraph(hierarchicElement);

		KShapeLayout par = parentNode.getData(KShapeLayout.class);
		par.setProperty(LayoutOptions.DIRECTION, Direction.RIGHT);
		par.setProperty(LayoutOptions.INTERACTIVE, true);
		par.setProperty(Properties.FEEDBACK_EDGES, true);
		par.setProperty(Properties.INTERACTIVE_REFERENCE_POINT, InteractiveReferencePoint.TOP_LEFT);

		AbstractLayoutProvider layoutProvider = new LayeredLayoutProvider();
		IKielerProgressMonitor progressMonitor = new BasicProgressMonitor();
		layoutProvider.doLayout(parentNode, progressMonitor);

		applyLayout(parentNode, hierarchicElement);

		AnnotationViewPartBase.setUpdateEnabled(avUpdateEnabled);
	}

	/**
	 * Rounds a given {@code float} value and decreases the value until it is a multiple of the grid
	 * size.
	 * 
	 * @param f
	 *            {@code float} to process
	 * 
	 * @return Given float value, rounded and aligned to the underlying grid.
	 */
	private int roundSnap2Grid(float f) {
		return roundSnap2Grid(f, false);
	}

	/**
	 * Rounds a given {@code float} value and aligns it the grid size.
	 * 
	 * @param f
	 *            {@code float} to process
	 * 
	 * @param enlarge
	 *            Whether to enlarge the given value while aligning to the grid.
	 * 
	 * @return Given float value, rounded and aligned to the underlying grid.
	 */
	private int roundSnap2Grid(float f, boolean enlarge) {
		int rval = (int)f;

		final int deviationFromGrid = rval % DEFAULT_GRID_SIZE;
		if(deviationFromGrid > 0)
			if(enlarge) {
				rval += DEFAULT_GRID_SIZE;
			}
		rval -= deviationFromGrid;

		return rval;
	}

	/**
	 * Applies the layout defined in the {@link KNode} follows to given {@link IHierarchicElement}.
	 * 
	 * @param template
	 *            The {@link KNode} that stores the layout.
	 * @param hierarchicElement
	 *            The {@link IHierarchicElement} to which the layout should be applied.
	 */
	private void applyLayout(KNode template, IHierarchicElement hierarchicElement) {
		ITopLevelElement topLevelElement =
				IPersistencyService.getInstance().getTopLevelElementFor(hierarchicElement);
		topLevelElement.runAsCommand(() -> {
			filterBendPoints(hierarchicElement);

			for(IHierarchicElement curElement : hierarchicElement.getContainedElements()) {
				KNode curNode = modelElementsToKNodes.get(curElement);
				KShapeLayout curLayout = curNode.getData(KShapeLayout.class);
				LayoutDataUtils.setNodeLayoutData((ILayoutedModelElement)curElement,
						roundSnap2Grid(curLayout.getXpos()), roundSnap2Grid(curLayout.getYpos()),
						roundSnap2Grid(curLayout.getWidth(), true),
						roundSnap2Grid(curLayout.getHeight(), true));
			}

			for(Entry<IConnection, KEdge> curEntry : connectionsToKEdges) {
				KEdge curEdge = curEntry.getValue();
				IConnection curConnection = curEntry.getKey();
				KEdgeLayout curLayout = curEdge.getData(KEdgeLayout.class);
				for(KPoint k : curLayout.getBendPoints()) {
					LayoutDataUIUtils.addBendPointToConnection(
							(ILayoutedModelElement)curConnection, roundSnap2Grid(k.getX()),
							roundSnap2Grid(k.getY()));
				}
			}

			setPositionsForPorts(outboundConnectorsToKPorts);
			setPositionsForPorts(inboundConnectorsToKPorts);
			setPositionsForGlobalPorts(globalInboundConnectorsToKPorts);
			setPositionsForGlobalPorts(globalOutboundConnectorsToKPorts);
		});
	}

	/**
	 * Applies the layout of the specified {@link KPort}s to their matching (global)
	 * {@link IConnector}s. Global means in this case that the {@link IConnector}s do not have a
	 * corresponding node, i.e. because they contain connections to a higher-level model.
	 * 
	 * @param globalPorts
	 *            A bijection from {@link IConnector}s to {@link KPort}s
	 */
	private void setPositionsForGlobalPorts(EMap<IConnector, KPort> globalPorts) {
		for(Entry<IConnector, KPort> i : globalPorts) {
			IConnector port = i.getKey();
			KPort p = i.getValue();
			KNode temp = p.getNode();
			KShapeLayout lay = temp.getData(KShapeLayout.class);
			setNodePosition((ILayoutedModelElement)port, roundSnap2Grid(lay.getXpos()),
					roundSnap2Grid(lay.getYpos()));
		}
	}

	/**
	 * Applies the layout of the specified {@link KPort}s to their matching {@link IConnector}s. The
	 * {@link IConnector}s must have corresponding nodes, i.e. they can not contain an
	 * {@link IConnection} to a higher-level model.
	 * 
	 * @param ports
	 *            A bijection from {@link IConnector}s to {@link KPort}s.
	 */
	private void setPositionsForPorts(EMap<IConnector, KPort> ports) {
		for(Entry<IConnector, KPort> ic : ports) {
			KPort kp = ic.getValue();
			IConnector port = ic.getKey();
			KShapeLayout portLayout = kp.getData(KShapeLayout.class);
			PortSide side = portLayout.getProperty(LayoutOptions.PORT_SIDE);
			EOrientation orientation = null;
			switch(side) {
				case NORTH:
					orientation = EOrientation.NORTH;
					break;
				case WEST:
					orientation = EOrientation.WEST;
					break;
				case EAST:
					orientation = EOrientation.EAST;
					break;
				case SOUTH:
					orientation = EOrientation.SOUTH;
					break;
			}

			setStickyConnectorLayoutData((ILayoutedModelElement)port, orientation,
					roundSnap2Grid((portLayout.getYpos())));
		}
	}

	/**
	 * Removes all bend-points in connections contained in the given {@link IHierarchicElement}.
	 * 
	 * @param hierarchicElement
	 */
	private void filterBendPoints(IHierarchicElement hierarchicElement) {
		EList<IConnection> clears = hierarchicElement.getConnections();
		for(IConnection c : clears) {
			EList<ILayoutData> l = ((ILayoutedModelElement)c).getLayoutData();
			for(ILayoutData i : l) {
				if(i instanceof Points || i instanceof PointsImpl) {
					((Points)i).getPoints().clear();
				}
			}
		}
	}

	/**
	 * Creates a KIELER {@link KNode} graph with the same structure as the given
	 * {@link IHierarchicElement}.
	 * 
	 * @param hierarchicElement
	 *            The {@link IHierarchicElement} for which the {@link KNode} graph should be created
	 * 
	 * @return A {@link KNode} graph with the same structure as the graph contained in the
	 *         {@link IHierarchicElement}.
	 */
	private KNode createKIELERGraph(IHierarchicElement hierarchicElement) {

		KNode parentNode = KimlUtil.createInitializedNode();

		modelElementsToKNodes = new BasicEMap<IHierarchicElement, KNode>();
		outboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();
		inboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();
		connectionsToKEdges = new BasicEMap<IConnection, KEdge>();
		undirectedConnectorsToKPorts = new BasicEMap<IConnector, KPort>();

		// create nodes
		for(IHierarchicElement curElement : hierarchicElement.getContainedElements()) {

			if(!(curElement instanceof ILayoutedModelElement)) {
				continue;
			}

			KNode curKNode = KimlUtil.createInitializedNode();

			KShapeLayout curLayout = curKNode.getData(KShapeLayout.class);

			EList<IConnector> inputPorts = new BasicEList<IConnector>();
			EList<IConnector> outputPorts = new BasicEList<IConnector>();
			EList<IConnector> genericPorts = new BasicEList<IConnector>();

			EList<IConnector> ports = curElement.getConnectors();
			for(IConnector curConnector : ports) {
				if(curConnector instanceof EntryConnectorBase) {
					inputPorts.add(curConnector);
				} else if(curConnector instanceof ExitConnectorBase) {
					outputPorts.add(curConnector);
				} else {
					outputPorts.add(curConnector);
					genericPorts.add(curConnector);
				}
			}

			Dimension dimension = getNodeSize((ILayoutedModelElement)curElement);

			int portNmbr = Math.max(inputPorts.size(), outputPorts.size());
			curLayout.setHeight(Math.max((1 + 2 * portNmbr) * DEFAULT_CONNECTOR_SIZE,
					roundSnap2Grid(dimension.getHeight(), true)));
			curLayout.setWidth(Math.max(DEFAULT_SHAPE_MINIMUM_HEIGHT,
					roundSnap2Grid(dimension.getWidth(), true)));

			curKNode.setParent(parentNode);
			modelElementsToKNodes.put(curElement, curKNode);
			for(IConnector curPort : outputPorts) {
				KPort k = createKPortFromIConnector(curPort, curKNode, outboundConnectorsToKPorts);
				if(genericPorts.contains(curPort)) {
					undirectedConnectorsToKPorts.put(curPort, k);
				}
			}
			for(IConnector curPort : inputPorts) {
				createKPortFromIConnector(curPort, curKNode, inboundConnectorsToKPorts);
			}
		}

		// create virtual nodes for global ports
		globalInboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();
		globalOutboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();

		for(IConnector curPort : hierarchicElement.getConnectors()) {
			KNode virtual = KimlUtil.createInitializedNode();
			KShapeLayout virtualLayout = virtual.getData(KShapeLayout.class);
			virtualLayout.setHeight(DEFAULT_CONNECTOR_SIZE);
			virtualLayout.setWidth(2 * DEFAULT_CONNECTOR_SIZE);
			virtual.setParent(parentNode);

			KPort k = KimlUtil.createInitializedPort();
			k.setNode(virtual);
			if(curPort instanceof ExitConnectorBase) {
				globalOutboundConnectorsToKPorts.put(curPort, k);
			} else {
				globalInboundConnectorsToKPorts.put(curPort, k);
				outboundConnectorsToKPorts.put(curPort, k);
			}
		}

		// create edges
		for(Entry<IConnector, KPort> out : outboundConnectorsToKPorts) {
			for(IConnection ic : out.getKey().getOutgoing()) {
				KEdge curEdge = KimlUtil.createInitializedEdge();
				KPort sourceport = out.getValue();

				curEdge.setSource(sourceport.getNode());
				curEdge.setSourcePort(sourceport);
				sourceport.getEdges().add(curEdge);
				KPort targetport = inboundConnectorsToKPorts.get(ic.getTarget());
				if(targetport == null) {
					targetport = globalOutboundConnectorsToKPorts.get(ic.getTarget());
				}
				if(targetport == null) {
					targetport = undirectedConnectorsToKPorts.get(ic.getTarget());
				}
				curEdge.setTargetPort(targetport);
				curEdge.setTarget(targetport.getNode());
				targetport.getEdges().add(curEdge);
				connectionsToKEdges.put(ic, curEdge);
			}
		}
		return parentNode;
	}

	/**
	 * Converts a given {@link IConnection} into a {@link KPort}, and registers it with the given
	 * {@link KNode} (representing the {@link IHierarchicElement} containing the {@link IConnector}
	 * ) and a {@link IConnector}-to- {@link KPort} map.
	 * 
	 * @param connector
	 *            {@link IConnector} to be converted to {@link KPort}
	 * @param knode
	 *            {@link KNode} that will contain the newly created {@link KPort}
	 * @param connectorsToKPorts
	 *            Map between {@link IConnector}s and {@link KPort}s.
	 * 
	 * @return {@link KNode} represeting the given {@link IConnector}.
	 */
	private KPort createKPortFromIConnector(IConnector connector, KNode knode,
			EMap<IConnector, KPort> connectorsToKPorts) {
		KPort k = KimlUtil.createInitializedPort();
		k.setNode(knode);
		KShapeLayout portLayout = k.getData(KShapeLayout.class);
		portLayout.setHeight(DEFAULT_CONNECTOR_SIZE);

		connectorsToKPorts.put(connector, k);
		return k;
	}
}
